#! /bin/gawk  -f

# ----------------------------------------------------------------------------
# Copyright (C) 1989, 1991-2020 Free Software Foundation.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see http://www.gnu.org/licenses/.
# ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
# Developers:
#   Jannick
#   A. Schwarz
# ----------------------------------------------------------------------------



# ----------------------------------------------------------------------------
#
# SYNOPSIS
#     mkpage.sh [-FSEP] [-v site=URL] [-v name=name] [comment.txt] file1.csv ...
#
# DESCRIPTION
#     -f    gawk script to be executed
#     -FSEP Sincle character field seperator used in csv file 
#           SEP can not be a double quote (")
#           SEP must be a single character
#           default is comma (,)
#     .csv  field separated file
#           title SEP author SEB target SEP URL
#           title    book title or table topic
#           author   book author or null
#           target   if present, target="_blank" is generated
#           URL      hyperlink to downloadable address of book or null
#     -v site=URL
#           URL where the file_TABLES html file is located. For example,
#              -v site=https://slipbits.com
#           The generated base file will contain references to the
#           file_TABLES file as in "<a href="URL/file_TABLES.html">.
#     -v name=name
#           Output file name as in name.html and name_TABLES.
#           If name is not given, then the output file name begins
#           with the file name without the '.csv' suffix.
#     comment.txt A comment file inserted in name.html
#
# OUTPUT FILES
#  -v name=name supplied
#     namt.html         Contains references to tables in file1_TABLES
#     name_TABLES       Contains a table of contents and tables for each topic
#  -v not supplied
#     filex.html        Contains a reference table to all topics in filex.csv
#
# ----------------------------------------------------------------------------
#
# The Problem:
#  I have a website (slipbits.com). Some of my pages reference pages which have a
#  table whose entries reference related items. For example, there are 20+ math
#  categories in my math page. Each category references another page with a single
#  table. After a tad I realized that since all the references were related (all math
#  topics) then I could replace the 20+ pages with a single page containing
#  tables for each related item, and reference those tables in the math page.
#
#  That is, replace:
#  main page   topic page with a single table
#     math->Abstract Algebra  # contains a single table
#     math->...               # contains a single table
#     math->Vector Analysis   # contains a single table
#
#  with:
#  main page   single page with multiple tables
#     math->math_TABLES#Abstract Algebra  # reference to a table in math_TABLES
#     math->...                           # reference to a table in math_TABLES
#     math->math_TABLES#Vector Analysis   # reference to a table in math_TABLES
#
#  This reduces the total number of pages on my Wordpress site and automates the
#  process for creating pages. And, it can be used throughout to reduce the number
#  of distinct pages required.
#
#  So hey, why not I says.
# ----------------------------------------------------------------------------
#  And then we get two issues (and I thought I was done). So suppose we have
#  several .csv files and each file has one or more tables. What we would
#  like to do is create a single reference file and in that reference file
#  group the tables and then reference them. Well, we can to that. Let's
#  look at what that means.
#
#  Suppose we have three files.
#  Filename                tables
#  Knots.csv       {macrame   knot-tying local_clubs}
#  Kites.csv       {box_kites war_kites  weather_kites}
#  Knitting.csv    {how_to    cupplies   local_clubs}
#
#  -v name="" or no name= is supplied
#
#  mkpage -F, Knots Kites Knitting  # which yields
#                             |                           |
#              Knot_TABLES    |           Kites_TABLES    |            Knitting_TABLES
#           o-- macrame       |          o --box_kites    |             o-- how_to
#           |                 |          |                |             |
#   Knots --o-- knot_tying    |  Kites --o-- war_kites    |  Knitting --o-- cupplies
#           |                 |          |                |             |
#           o-- local_clubs   |          o-- weather_kites|             o-- local_clubs
#
#  Two separate pairs of HTML files are created for each input file. One file has
#  references to tables in located in the second file. The intent is to provide
#  a user seeable web page which has hyperlinks to tables is a separate file.
#
#  It looks a little different if -v name=text
#
#  mkpage -F, -v name=Leisure Knots Kites Knitting  # which yields
#
#                           Leisure_TABLES
#                           o-- macrame
#                           |
#             o-- Knots ----o-- knot_tying
#             |             |
#             |             o-- local_clubs
#             |
#             |             o-- box_kites
#             |             |
#   Leisure --o-- Kites ----o-- war_kites
#             |             |
#             |             o-- weater_kites
#             |
#             |             o-- how_to
#             |             |
#             o-- Knitting--o-- supplies
#                           |
#                           o-- local_clubs
#
#  Now we have two files. One file contains a reference to tables in the _TABLES file, and
#  the _TABLES file has the concatenation of all tables from all the input files.
#
#  Table names are ignored, other than to provide this name as a caption for a table.
#  This means that either in the same file or across different files, the same table
#  name can be used without conflict.
#
# ----------------------------------------------------------------------------
#
# Restriction:
#  This code is meant to process spreadsheet csv (comma separated list) output.
#  The csv file is an ASCII text file with comma's separating each cell and a
#  line feed <lf> character terminating a line. Gawk uses the comma's to determine
#  input fields (each cell is called a field in gawk-land), and the <lf> character
#  to determine a line (a row in spreadsheet-land).
#
#  The translation(s) look like:
#           Spreadsheet                       .csv file
#     row: cell1,cell2,cell3  === csv === cell1,cell2,cell3<lf>
#
#          .csv file                         gawk fields
#     cell1,cell2,cell3<lf>   === gawk == f1 f2 f3
#
#  If a spreadsheet cell has a comma (,) in it, then the .csv output is:
#
#     spreadsheet              .csv file
#        x,y         === csv === "x,y"    # "x,y" is processed by gawk
#
#  if a spreadsheet cell has a double quote (") in it:
#
#     spreadsheet              .csv file
#        x"y         === csv === x""y    # "x,y" is processed by gawk
#
#  When the double quote is the first character of a cell, or if the cell has
#  both a double quote and a comma, it gets confusing:
#
#         "x y  === .csv === ""x,y
#         "x, y === .csv === """x,",y
#
#  Linefeeds (<lf> or \n) act as line terminators in gawk. Each time a <lf> is seen,
#  gawk creates a new input line. For example:
#
#               Spreadsheet                            .csv file
#     row: ti\ntle author target URL === csv === ti\ntle,author,target,URL<lf>
#
#
#                                                o- ti                         line1
#                                                |
#                .csv file                       |
#     ti\ntle,author,target,URL<lf>  === gawk ===o
#                                                |
#                                                |
#                                                o- tle, author, target, URL   line2
#
#
#  Restriction:
#  1. If double quotes are used then the double quotes have the doubled (" => "").
#     and the field containing the double quotes quoted (a"b => "a""b").
#` 2. If the separator character is used in a field then the field must be quoted,
#     (a,b => "a,b" if comma (,) is the separator character).
#  3. Do not use linefeeds in a field.
#  4. Do not use a carriage return as a line terminator.
#
# ----------------------------------------------------------------------------
#
# Comments
#  The user can insert three comments. All comments are optional
#  1: The file containing references.
#     a. Each reference can have a one-line comment.
#  2: Each table can have a preceding comment describing the table.
#
#  This files look like:
#
#            reference file
#  ---------------------------------------
#  | text describing the reference file, |
#  | the general topic are, and most     |
#  | anything else.                      |
#  |                                     |
#  | hyperlinked table comment           |
#  |             o o o                   |
#  | hyperlinked table comment           |
#  ---------------------------------------
#  |
#
#              table file
#  ---------------------------------------
#  | Table of Contents                  |
#  |                   |
#  | table comment text                  |
#  |        o o o                        |
#  | table comment text                  |
#  | table caption                       |
#  | hyperlinked book title | author     |
#  |        o o o                        |
#  | hyperlinked book title | author     |
#  |                                     |
#  |        o o o                        |
#  |                                     |
#  | table comment text                  |
#  | table caption                       |
#  | hyperlinked book title | author     |
#  |        o o o                        |
#  | hyperlinked book title | author     |
#  |                                     |
#  ---------------------------------------
#
#  The three types of comments are input as:
#  1: reference file: column 1 of the .csv file has a "#!" in it.
#     Column 2 contains the input text. Each new row becomes a
#     new paragrpah unless the row is an HTML tag (<tag>) in
#     which case it becomes a new physical line with the tag
#     and any following text.
#  2: table file: column 1 of the .csv file has a "#" in it.
#     Column 2 contains the input text which is treated as
#     above. Table comments must immediately precede the
#     table caption in the .csv file.
#  3: hyperlinked table comment: any text in column 5 of the
#     input .csv file is consiered as a one line comment.
#     Column 5 content is only evaluated for table captions
#
# Restrictions
#  1. Do not use double quotes.
#  2, Do not use Line feeds.
#
# ----------------------------------------------------------------------------
#
# Input .csv Format:
#  A .csv (csv output of a spreadsheet) file with four fields:
#     caption,,,comment  or
#     title,author,target,URL
#     #,! input_file text
#     #,  table comments
#
#  where
#     caption  table caption <caption>caption</caption>
#     title    book title
#     author   book author
#     target   "" if the book is directly downloadable     <a href="">
#              [:print] if there is a required interaction <a href="" target="_blank">
#     URL      hyperlink address of the book download
#     comment  One-liner to be inserted in the reference to the table
#
#  All fields can be double quoted, as in "caption"
#  Any field with a bare comma (',') must be double quoted.
#  Any field with a double quote ('"') must be double quoted where the internal double quote must be doubled.
#
# ----------------------------------------------------------------------------
#
# Output files
#  -v name=name
#     name.html        reference file to tables
#     name_TABLES      table file
#
#  otherwise
#     filename.html    Comprehensive file containing both references to tables and tables
#
# ----------------------------------------------------------------------------
#
# Processing assumptions:
#
#  if title is present and author = target = URL = <null> then this is a caption
#  and terminates the previous table.
#
#  if URL is present then name must be present
#  if name = author = target = URL = <null> then this terminates a topic
#
# Legal input lines
#  caption,,,comment          new topic              <a name=""><table></a><caption><b>caption</b></caption>
#  title,,,URL                table entry w/o author <a href="URL">
#  title,,x,URL               table entry w/o author <a href="URL" target="_blank">
#  title,author,,URL          table entry w/ author  <a href="URL">
#  title,author,x,URL         table entry w/ author  <a href="URL" target="_blank">
#  ,,,,                       topic terminator (optional) </table>
#
# ----------------------------------------------------------------------------
#
# Generated text (author may be <null>)
#  new topic:   <a name="__TableRef___Toc####">table name</a>
#                 <caption><b>text</b></caption>
#                    <tr>
#                       <th  width="396px"><center>title</center><;th>
#                       <th>author</th>
#                    </tr>
#  new topic:   <h1 name="__TableRef___Toc####"</h1>
#                 Table comments
#                 <table>
#                 <caption><b>text</b></caption>
#
#  table entry:      <tr>
#                       <td><a href="URL">title</a></td>
#                       <td>author</td>
#                    </tr>
#  table entry:      <tr>
#                       <td><a href="URL" target="_blank">title></a></td>
#                       <td>author</td>
#                    </tr>
#  topic end:   </table>
#
#
# PASSING INPUT ARGUMENTS: https://stackoverflow.com/questions/19075671/how-do-i-use-shell-variables-in-an-awk-script
#
# ----------------------------------------------------------------------------
#  THEORY OF OPERATIONS:
#
#  This is a multipass process.
#  1: Initialization:
#     a. Statically create all strings used during processing.
#     b. Initialize all transient variables.
#
#  2: BEGINFILE processing: Initialize variables for the next file
#     a. Determine the output file name and website URL.
#     b. Initialize all variables.
#
#  3: File processing pass 1:
#     a. If the input is a comment file ("file.txt") then process it and get next file.
#     b. Otherwise, collect table data and table content data.
#  4: ENDFILE processing:
#     a. If this is single table generateion, then
#        1. Generate the single web page.
#        2. Include comments.
#        3. Include a table header.
#        4. Include table entries.
#        5. Generate web page table/page termination.
#     b. If this is a multi-table web page, do nothing.
#  5: END processing:
#     a. If this is a single table to be generated, to nothing.
#     b. If this is a multi-table web page.
#        1. Generate the reference web page table references.
#        2. Generate the table (name_TABLES) web page.
#        3. Include a table header for each table.
#        4. Include table entries for each table.
#        5. Generate table termination for each table.
#        5. Generate table web page termination.
#
#
# ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
# Initialization
#
#   Initialize global constants and variables.
#   reHead        HTML for the file. Establishes DOCTYPE, styles, and beginning of <head> field.
#   TOCLine       HTML for Table of Contents.
#   TOCEnd        HTML for the end of the Table of Content <div>.
#   bodyEnd       HTML for <body> end and document end.
#   h1Header      HTML for <h1>, used as TOC reference.
#   tableCaption  HTML for the start of a <table>.
#   bookColumn    HTML for the book name entry in a <table>.
#   authorColumn  HTML for the author entry in a <table>.
#   baseHead      HTML for the start of the reference list entry to the <table>s.
#   baseLine      HTML for an entry in the reference list.
#   baseTitle     HTML for the name of a reference category.
#   bseEndList    HTML for the end of the reference list.
#   target        HTML for table entries which include "target".
#   bodyEnd       HTML for the end of a document.
#   startRef      Counter for reference entries.
# ----------------------------------------------------------------------------

BEGIN {                                         # program constants

        refHead[ 1] = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">"
        refHead[ 2] = "<html>"
        refHead[ 3] = "    <head>"
        refHead[ 4] = "       <meta http-equiv=\"content-type\" content=\"text/html; charset=windows-1252\"/>"
        refHead[ 5] = "       <title></title>"
        refHead[ 6] = "       <style type=\"text/css\">p.leaders{max-width:18cm;padding:0;overflow-x:hidden;line-height:120%}p.leaders:after{float:left;width:0;white-space:nowrap;content:\". . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . \"}p.leaders span:first-child{padding-right:0.33em;background:white}p.leaders span+span{float:right;padding-left:0.33em;background:white;position:relative;z-index:1}"
        refHead[ 7] = "            @page { size: 8.5in 11in; margin: 0.79in }"
        refHead[ 8] = "            a:link { color: #000080; so-language: zxx; text-decoration: underline }"
        refHead[ 9] = "            a:visited { color: #800000; so-language: zxx; text-decoration: underline }"
        refHead[10] = "            h1 { margin-bottom: 0.08in; background: transparent; page-break-before: auto; page-break-after: avoid }"
        refHead[11] = "            h1.western { font-family: \"Liberation Serif\", serif; font-size: 12pt; font-weight: bold }"
        refHead[12] = "            th.western { font-family: \"Liberation Serif\", serif; font-size: 12pt; font-weight: bold }"
        refHead[13] = "            caption.western { font-family: \"Liberation Serif\", serif; font-size: 12pt; font-weight: bold }"
        refHead[14] = "       </style>"
        refHead[15] = "    </head>"
        refHead[16] = "    <body lang=\"en-US\" link=\"#000080\" vlink=\"#800000\" dir=\"ltr\">"
        refHead[17] = "        <div id=\"Table of Contents1\" dir=\"ltr\">"
        refHead[18] = "            <div id=\"Table of Contents1_Head\" dir=\"ltr\"><p style=\"margin-top: 0.17in; margin-bottom: 0.08in; line-height: 100%; page-break-after: avoid\">"
        refHead[19] = "                <font face=\"Liberation Serif, serif\"><font size=\"4\" style=\"font-size: 16pt\"><b>Table of Contents</b></font></p>"
        refHead[20] = "            </div>"
        refHeadCount= 20;

        TOCLine     = "            <p class=\"leaders\"><span><font face=\"Liberation Serif, serif\"><font size=\"3\" style=\"font-size: 12pt\"><a href=\"#__TableRef__Toc%d\"></font>%s</a></span></p>\n"
        TOCEnd      = "        </div>"

        bodyEnd[1]  = "    </body>"
        bodyEnd[2]  = "</html>"

        h1Header    = "<h1 class=\"western\"><a name=\"__TableRef__Toc%s\"></a>\n%s\n    <table>\n"
        tableCaption= "    <caption class=\"western\"><b>%s</caption>\n"
        bookColumn  = "    <tr>\n          <th class=\"western\" width=\"500px\"><center>Book Title</center></th>\n"
        authorColumn= "          <th class=\"western\">Author</th></b>\n       </tr>\n"

        bsseHead    = "<ul border=\"0\">";
        baseLine    = "   %s<li><a href=\"%s%s#__TableRef__Toc%d\" target=\"_blank\">%s</a> %s</li>\n";
        baseTitle   = "   <li>%s</li>\n   <ul>\n";
        baseEndList = "   </ul>\n";
        target      = " target=\"_blank\">%s</a>%s</li>\n";
        baseEnd     = "</ul>"
        
#--------------------------------------------------# Initialize GAWK variables
        IGNORECASE  = 1                            # ignore case in all string comparison operations (incl. regexp matches)
        RS = "\r?\n"                               # record separator - either LF or CRLF 
        
        #
        # Processs user supplied SEP (-FSEP)
        
        if(!SEP) SEP = ",";
        
        else if (length(SEP) > 1) {
           printf "Value of command line argument SEP should be one single character, given '%s'\n", SEP > "/dev/stderr"
           exit 2
           
        } else if (SEP == "\"" ) {
           printf "Command line argument SEP must not be a double quote.\n" > "/dev/stderr"
           exit 2
        }
        
        FPAT         =  sprintf("(\"([^\"]|\"\")+\"|[^%s\"]*)",SEP)  # CSV field pattern


#--------------------------------------------------# Initialize program variables
       lineNo       = 0;                           # caption/table entry line count
       tableNdx     = 0;                           # list of tables start #
       tableCount   = 0;                           # total number of tables seen
       startRef     = 1000                         # start value for table reference
       tableRef     = startRef;                    # setup reference number

        if (name) {                                # -v name=text on command line
            baseFile= name;
            dirFile = name".html"
            refFile = name"_TABLES";
         }

} # BEGIN

# ----------------------------------------------------------------------------
# Initializes required values at the start of each file.
#
# Before a file is processed various values required by the file are
# initialized.
# ----------------------------------------------------------------------------
BEGINFILE {                                        # file based Globle Variables
        end      = 1;                              # output </table> toggle
        line     = "";                             # input line
        qnum     = 0;                              # number of quotes in a line
        tableRef = (name)?tableRef :startRef;      # start num of name=__TableRef__Tocxxx
        text     = ""                              # comments for baseFile and before refFile tables
        tableDescription = ""                      # decription precedes table
                                                   #            input line template
        title    = "";                             # title:     title,,,,
        author   = "";                             # author:    ,author,,,
        target   = "";                             # target:    ,,target,,
        URL      = "";                             # hyperlink: ,,,URL,
        comment  = "";                             # comment    ,,,,comment
        
        if (site) {
            sub(/\/+$/, "", site);
            site     = site "/";                   # URL/
        }

        if (name) {                                # -v name=name on command line
            savedFileName= deleteExtension();
            gsub(/\-/, " ", savedFileName);
        } else {
           lineNo     = 0;                         # non-null line count
           tableCount = 0;                         # total number of tables seen
           savedFileName = "";                     # acts as a flag in outputBaseRef
           delete tableList;                       # list of tables

           if(baseFile = deleteExtension()) {      # get base name for file
               dirFile = baseFile".html"
               refFile = baseFile"_TABLES.html";   # file_TABLES
           } else {
               outputError("EXIT: " FILENAME " not legal");
               exit;
           }
        }
   if(dbg)  print FILENAME, " <baseFile,refFile> <", baseFile, ", ", refFile, ">\n";        # debug
 } # BEGINFILE

# ----------------------------------------------------------------------------
# Comment line processing.
#
# Each line is trimmed (no leading/trailing space) and any leading/trailing
# double quotes are removed (").
# ----------------------------------------------------------------------------
function commentProcessing(comment) {              # process double quotes in comments
   cmnt = trim(comment);

   if (cmnt == "") return cmnt;                    # blank line

   sub(/^"/, "", cmnt);                            # remove leading double quote
   sub(/"$/, "", cmnt);                            # remove trailing double quote
   if (cmnt !~/^[[:space:]]*</)                    # if ho leading HTML tag
      cmnt = "<p>" cmnt "</p>";
   return cmnt;
} # function commentProcessing(comment)

# ----------------------------------------------------------------------------
#  Remove extension from files (name.csv). The nmae becomes the table
#  reference name in the reference file.
#
#  Return:
#     filename.txt  returned for filename.txt if -v name=filename
#     filename      returned for filename.csv
# ----------------------------------------------------------------------------
### J: better to use arguments passed on to the function to keep the number of global variables as small as possible.
function deleteExtension(   array, i, myFile, num) {                       # delete file extension (file.ext => file)
   myFile = "";

   if (FILENAME == name".txt")   # comment file
      return FILENAME;

   num = split(FILENAME, array, /\./);

   if  (num < 2) {                                 # no extension
      outputError(FILENAME " missing .csv extension");
   } else if (array[num] != "csv") {               # extension not .csv
      outputError(FILENAME " extension must be .csv");
   } else {
      #for(i = 1; i < num; i++) myFile = myFile array[i]; # Are the dots missing as delimiters?
      myFile = FILENAME
      gsub(/\.[^.]+$/,"", myFile)
   }
   return myFile;
} # function deleteExtension()

# ----------------------------------------------------------------------------
#  At the end of input for each file provide HTML to end the created table(s).
# ----------------------------------------------------------------------------
function endfileProcessing() {               # complete end of file/named group processing
   cnt    = lineNo;                          # preserve total count of input items
   tableCount = tableNdx;                    # total number of tables

   if(dbg) print "baseFile: ", dirFile, "\n\n";     # debug
   print bsseHead                            > dirFile; # output references to refFile
   (name)? outputNamedRef():outputBaseRef();
   print baseEnd                             > dirFile;

   outputRefHeader(cnt);
   outputTOC(cnt);
   print TOCEnd                              > refFile;

   lineNo = 0;                               # reset global line number
   for( i = 1; i <= cnt; i++) pass2();       # process table entries

   print bodyEnd[1], "\n", bodyEnd[2]        > refFile;
   delete fieldValue
} # function endfileProcessing()


# ----------------------------------------------------------------------------
#  Output a table entry into the table file.
#
# ----------------------------------------------------------------------------
function entry() {                                 # output table entry
   printf("       <tr>\n")  > refFile;             # <td><a href="URL" [target="_blank"]>title</a></td>
   printf("          <td><a href=\"%s\"", URL)           > refFile;
   if (target) { printf(" target=\"_blank\"")            > refFile; }
   printf(">%s</a></td>\n", title)                       > refFile;
   printf("          <td>%s</td>\n", author)             > refFile;
   printf("       </tr>\n")                              > refFile;
} # function entry()

# ----------------------------------------------------------------------------
#  Output a reference to a table in the Table of Contents in a table file.
#
# ----------------------------------------------------------------------------
function header(topic, comment) {                  # output table header
   printf(h1Header, tableRef, comment)                   > refFile;
   printf(tableCaption, topic)                           > refFile;
   printf(bookColumn)                                    > refFile;
   printf(authorColumn)                                  > refFile;
   printf(tableHeader)                                   > refFile;
} # function header(topic, comment)

# ----------------------------------------------------------------------------
#  Determine if the current input line is a caption.
#
#  return 1 input field is a caption
# ----------------------------------------------------------------------------
function isCaption(ndx) {                          # is this a table caption
   tableFlag = 0;
   if (fieldValue[ndx][1] &&                       # 'title' is non-null
      (length(fieldValue[ndx][4]) < 2))            # URL null
         tableFlag = 1;
   return tableFlag;
} # function isCaption()

# ----------------------------------------------------------------------------
#  Report errors.
# ----------------------------------------------------------------------------
function outputError(msg) {                        # output diagnostic message
   print msg > "/dev/stderr"
} # function outputError(msg)

# ----------------------------------------------------------------------------
#  Output the header HTML for the Table of Contents into the table file.
#
# ----------------------------------------------------------------------------
function outputRefHeader(lineNo) {                 # output HTML stuff
  for ( i = 1; i <= refHeadCount; i++) print refHead[i]         > refFile;
  if (lineNo) print "<p><center><b>Table of Contents</b></center></p>" > refFile;
} # function outputRefHeader(lineNo)

# ----------------------------------------------------------------------------
#  Generate table references in the table reference file (name.html).
# ----------------------------------------------------------------------------
function outputBaseRef() {                         # output references to baseFile
   tableNdx   = 0;                                 # reset index to start
   for( ; tableNdx < tableCount;) {                # process table entries
      getTable();                                  # side-effect to increment tableNdx
      printf(baseLine, prefix, site, refFile, tableRef, title, comment) > dirFile;
   }
} # function outputBaseRef()

# ----------------------------------------------------------------------------
# Output references to tables to the comment file, if one is provided, or to
# the reference file to the generated tables.
#
# ----------------------------------------------------------------------------
function outputNamedRef() {                        # output references to the named file
   prefix   = "";                                  # leading spaces before title output
   tableNdx   = 0;                                 # reset index to start
   tableList[tableCount+1][1] = "";                # setup guard to list
   tableList[tableCount+1][2] = "";
   tableList[tableCount+1][3] = "";

   for( ; tableNdx < tableCount;) {                # process table entries
      getTable();                                  # side-effect to increment tableNdx
      newFileName = tableList[tableNdx+1][2];
      if (fileName == newFileName) {               # Create sublist
         prefix = "   ";
         subListName = fileName;
         printf(baseTitle, fileName)                                       > dirFile;  # Output filename
         do {
            printf(baseLine, prefix, site, refFile, tableRef, title, comment) > dirFile;
            getTable();
            newFileName = tableList[tableNdx+1][2];
            if(dbg)   print "[", tableNdx, ", }", tableCount, "<", fileName, ", ", newFileName, ">" # debug
         } while (fileName == newFileName)
         printf(baseLine, prefix, site, refFile, tableRef, title, comment) > dirFile;
         printf(baseEndList) > dirFile;            # terminate previous group
         prefix = "";
      } else {
         printf(baseLine, prefix, site, refFile, tableRef, title, comment) > dirFile;
      }
   }
} # function outputNamedRef()

# ----------------------------------------------------------------------------
#  Output the table of contents to the table file ("name_TABLES").
# ----------------------------------------------------------------------------
function outputTOC(cnt) {                          # output Table of Contents to refFile
   lineNo      = 0;                                # reset global line number
   for( i = 1; i <= cnt; i++) {                    # process table entries
      getFields()
      if (isCaption(i)) {                          # 'author' is non-null
         printf(TOCLine, tableRef, title)             > refFile;
      }
   }
} # function outputTOC(cnt)

# ----------------------------------------------------------------------------
#  Extract the fields from the current input line.
#
#  There is a difference of opinion between the way that gawk treats input
#  lines and the way csv readers, most particularly from LibreOffice, treat
#  input. Gawk separates the input lines into fields based soley on the
#  separator character SEP, or the separator pattern, FPAT. csv readers are 
#  more nuanced. Input lines with a leading (and trailing) double quote are 
#  treated as a single field, cell in csv terminoloty, without regard to 
#  whether SEP character or FPAT pattern is within the field. This provides 
#  an issue. Suppose that a quoted field contains SEP. gawk will yield two 
#  fields but csv will only yield one.
#
#  For example:         gawk      csv
#                      fields    fields
#     "abc,def",g        abc     abc,def
#                        def     g
#                        g
#
#  FPAT takes eliminates this issue by accepting either a field contained in
#  double quotes, or a field without sorrounding double quotes.
#
#  For example:         gawk      csv
#                      fields    fields
#     "abc,def",g      abc,def   abc,def
#                      g         g
#
#  An embedded double quote must be doubled, that is " => "", to be an 
#  acceptable csv field. This jewel of a routine changes the double
#  quote to a single quote.
#
#  For example:         gawk      csv
#                      fields    fields
#    "abc""def",g      abc"def   abc,def
#                      g         g
#
#  Input Parsing of quoted lines converts csv cells to gawk fields.
#  Leading and trailing whitespace is removed from each field.
#
# ----------------------------------------------------------------------------
function fieldProcessing() {                       # fix field w/"  ,  "
   title = author = target = URL = "";
   
   for (i = 1; i <= NF; i++) {                     # fore each field in the input line
      gsub(/(^"|"$)/,"",$i)FPAT                    # remove leading/trailing double quote
      gsub(/""/,"\"",$i)                           # change "" to "
   }

   title  = trim($1);
   author = trim($2);
   target = trim($3);
   URL    = trim($4);
   comment= trim($5);


} # function fieldProcessing()

# ----------------------------------------------------------------------------
#  Retrieve all field information saved during pass1 for a given input line.
# ----------------------------------------------------------------------------
function getFields() {                             # restore fields
   lineNo++;
   title        = fieldValue[lineNo][1];
   author       = fieldValue[lineNo][2];
   target       = fieldValue[lineNo][3];
   URL          = fieldValue[lineNo][4];
   comment      = fieldValue[lineNo][5];
   tableDescription = fieldValue[lineNo][6];
   if (fieldValue[lineNo][7]) {
      if(dbg)  printf("get  FieldVlue[%4d] %4d, %s, %s, %s, %s\n",  lineNo,   tableRef, title, author, target, (URL)? "<li>": "caption"); # debug
      tableRef     = fieldValue[lineNo][7];
   }
} # function getFields()

# ----------------------------------------------------------------------------
#
# ----------------------------------------------------------------------------
function getTable() {
   tableNdx++;                                     # dequeue index
   title       = tableList[tableNdx][1];
   fileName    = tableList[tableNdx][2];
   comment     = tableList[tableNdx][3];
   tableRef    = tableList[tableNdx][4];
#   printf("get  tableList[%4d] %4d, %s, %s\n",          tableNdx, tableRef, title, savedFileName); # debug
}

# ----------------------------------------------------------------------------
# Save data required for pass 2.
#  
#  Pass 2 uses the data discoved in pass 1 to construct the output reference
#  file (name.html) and table file (name_TABLES).
#
#  It is required that a valid table caption (title,,,) precede the table contents.
#  
#  fieldValue array: Contains information on each input line (from csv files)
#                    for all files, and for tables, the reference number of the
#                    table.
#  tableList array:  Contains information on each table.
# ----------------------------------------------------------------------------
function saveFields() {                            # save fields for the second pass
   lineNo++;
   fieldValue[lineNo][1] = title;
   fieldValue[lineNo][2] = author;
   fieldValue[lineNo][3] = target;
   fieldValue[lineNo][4] = URL;
   fieldValue[lineNo][5] = comment;
   fieldValue[lineNo][7] = "";

   if (isCaption(lineNo)) {                        # gawks version of a structure
      tableNdx++;                                  # queue index
      tableList[tableNdx][1] = title;              # caption for table
      tableList[tableNdx][2] = savedFileName;      # .csv filename containining the table
      tableList[tableNdx][3] = tableDescription;   # table comment.
      tableList[tableNdx][4] = tableRef;
      fieldValue[lineNo][7]  = tableRef;
      if(dbg) {
         printf("save FieldVlue[%4d] %s, %s, %s, %s\n",  lineNo,  title, author, target, (URL)? "<li>": "caption");  #debug
         printf("     tableList[%4d] %4d, %s, %s\n",     tableNdx, tableRef, title, savedFileName); # debug
      }
      tableRef += 10;
   }
# debug
} # function saveFields()

# ----------------------------------------------------------------------------
#  Trim leading/trailing white space from the current input line.
# ----------------------------------------------------------------------------
function trim(field) {                             # trim whitespace from field
   sub(/^[[:space:]]*/, "", field);                # leading whitespace
   sub(/[[:space:]]*$/, "", field);                # trailing whitespace
   return field
} # function trim(field)

# ----------------------------------------------------------------------------
#  Process the input file:
#
#  Determine if processing is to create a comprehensive directory reference
#  file or if each input file is to generate a sole output file containing
#  both tables and reference to the tables.
#
#  If the comprehenisve directory is to be processed then if the current
#  input is a text file ("file.txt"), then process the text file and continue
#  to the next file. Otherwise, process the input ".csv" file.
#
#  Process the ".csv" file:
#     if the input line is blank, continue to the next line
#     otherwise, initialize the processing state (fields) and process table
#     comments and continue to the next line, or process the current line of
#     file. then continue to the next line.
# ----------------------------------------------------------------------------
function pass1() {                                 # collect topic, table and comments
   if(dbg) print "[", NR, "] => ",  $0;            # debug
   if (name) {
      if (FILENAME == baseFile".txt"){             # comment.txt file processing
         print commentProcessing($0)               > dirFile;
         next;
      }
   }

   if (NF) {                                       # not a blank line
      fieldProcessing();
      if (match(title, /^#/)) {                    # check which comment
         if (title == "#!") {
            if (!name) print comment               > dirFile;
         } else tableDescription = tableDescription "\n" comment;
         next;
      }
      saveFields();
      tableDescription = "";                       # reset table comment
   }
} # function pass1()

# ----------------------------------------------------------------------------
#  Generate a table header/trailor.
#
# ----------------------------------------------------------------------------
function pass2() {                                 # output TOC & tables
   getFields()
   if (title) {                                    # 'title' is non-null
      if (length(URL) > 1) {                       # URL non-null
         entry();
      } else  {                                    # URL is null => new topic
         if (end == 0) {                           # </table> was not generated
            printf("   </table>\n")          > refFile;
         }
         header(title, tableDescription)           # ourput table header
         end = 0;
      }
   } else {                                        # 'title is null => ",,,"
      if (end == 0) {                              # </table> output
         print("   </table>\n")              > refFile;
         end = 1;
      }
      if (author || target || length(URL) > 1) {   # error
         dec      = length(author);
         str      = sprintf("ILLEGAL INPUT: %s, %s, %s, %s", title, author, target, URL);
         outputError(str);
         if (author) { outputError("   no author allowed"); }
         if (target) { outputError("   no target allowed"); }
         if (length(URL) > 1) { outputError("   no URL allowed"); }
      }
   }
} # function pass2()

# ----------------------------------------------------------------------------
#  Process each line of the current file. If the current file is a comment
#  file ("file.txt"), then the entire file will be processed in 'pass1()'
#  and the next file will be retrieved, otherwise each line of the current
#  ".csv" file is handled here.
# ----------------------------------------------------------------------------

{
   pass1()
} # record sequencing loop

# ----------------------------------------------------------------------------
#
# ----------------------------------------------------------------------------
ENDFILE {
   if (!name) endfileProcessing();
} # ENDFILE

# ----------------------------------------------------------------------------
#
# ----------------------------------------------------------------------------
END {
   tableCount = tableNdx;
   if (name)  endfileProcessing();
}